--[[
Interface: 1.7.1.0 b10490

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 28.09.2020
Version: 1.0.0.0

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Free for use in other mods - no permission needed.
No changes are to be made to this script without permission from GtX | Andy

Frei verwendbar - keine erlaubnis nötig
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


RailroadBarrier = {}

RailroadBarrier.DEBUG_ACTIVE = false
RailroadBarrier.CHECK_TIME = 30000

local RailroadBarrier_mt = Class(RailroadBarrier)

function RailroadBarrier.onCreate(id)
    RailroadBarrier:createFromNode(id, g_currentMission.loadingMapBaseDirectory, g_currentMission.loadingMapModName)
end

function RailroadBarrier:createFromNode(id, baseDirectory, customEnvironment)
    local xmlFilename = getUserAttribute(id, "xmlFilename")

    if xmlFilename ~= nil then
        xmlFilename = Utils.getFilename(xmlFilename, baseDirectory)
    else
        g_logManager:error("Missing 'xmlFilename' user attribute for 'RailroadBarrier' node '%s'!", getName(id))
        return false
    end

    local index = getUserAttribute(id, "index")

    if index == nil then
        g_logManager:error("Missing 'index' user attribute for RailroadBarrier node '%s'!", getName(id))
        return false
    end

    local xmlFile = loadXMLFile("RailroadBarrier", xmlFilename)

    if xmlFile == 0 then
        return false
    end

    local baseKey = ""
    local i = 0

    while true do
        local objectKey = string.format("railroadBarriers.railroadBarrier(%d)", i)
        if not hasXMLProperty(xmlFile, objectKey) then
            break
        end

        local configIndex = getXMLString(xmlFile, objectKey .. "#index")

        if configIndex == index then
            baseKey = objectKey

            break
        end

        i = i + 1
    end

    if baseKey == nil then
        g_logManager:error("Index '%s' not found in xml file '%s'!", index, xmlFilename)
        return false
    end

    local instance = RailroadBarrier:new(g_server ~= nil, g_client ~= nil)

    if instance:load(id, xmlFile, baseKey, baseDirectory, customEnvironment, xmlFilename) then
        g_currentMission:addUpdateable(instance)
    else
        instance:delete()
    end

    delete(xmlFile)

    return true
end

function RailroadBarrier:new(isServer, isClient, customMt)
    local self = setmetatable({}, customMt or RailroadBarrier_mt)

    self.isServer = isServer
    self.isClient = isClient

    self.nodeId = 0

    self.animationDirection = 0

    self.lights = {}
    self.numLights = 0

    self.alternatingLights = {}
    self.numAlternatingLights = 0

    self.samples = nil
    self.samplesActive = false

    self.requiresUpdate = false

    self.objectsInTrigger = {}
    self.numObjectsInTrigger = 0

    self.activeTriggerTime = 0
    self.objectCheckTime = RailroadBarrier.CHECK_TIME

    return self
end

function RailroadBarrier:load(nodeId, xmlFile, key, baseDirectory, customEnvironment, xmlFilename)
    self.nodeId = nodeId
    self.identifierName = tostring(Utils.getNoNil(getUserAttribute(nodeId, "identifierName"), getName(nodeId)))

    self.baseDirectory = baseDirectory
    self.customEnvironment = g_currentMission.loadingMapModName

    self.triggerId = I3DUtil.indexToObject(nodeId, getXMLString(xmlFile, key .. ".trigger#node"))

    if self.triggerId ~= nil then
        local useRealLights = false
        local lightsProfile = g_gameSettings:getValue("lightsProfile")

        if lightsProfile == GS_PROFILE_HIGH or lightsProfile == GS_PROFILE_VERY_HIGH then
            useRealLights = true
        end

        addTrigger(self.triggerId, "triggerCallback", self)

        for i = 0, getNumOfChildren(self.triggerId) - 1 do
            addTrigger(getChildAt(self.triggerId, i), "triggerCallback", self)
        end

        self.animationParts = {}
        self.numAnimationParts = 0

        self.animationTime = 0
        self.animationDirection = 0
        self.animationDuration = math.max(Utils.getNoNil(getXMLFloat(xmlFile, key .. ".animation#duration"), 3), 0.1) * 1000

        local i = 0
        while true do
            local partKey = string.format("%s.animation.part(%d)", key, i)
            if not hasXMLProperty(xmlFile, partKey) then
                break
            end

            local node = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, partKey .. "#node"))

            if node ~= nil then
                local part = {
                    node = node,
                    animCurve = AnimCurve:new(linearInterpolatorN)
                }

                local hasFrames = false

                local j = 0
                while true do
                    local frameKey = string.format("%s.keyFrame(%d)", partKey, j)
                    if not hasXMLProperty(xmlFile, frameKey) then
                        break
                    end

                    local keyframe = {self:loadFrameValues(node, xmlFile, frameKey)}
                    keyframe.time = Utils.getNoNil(getXMLFloat(xmlFile, frameKey .. "#time"), 0)

                    part.animCurve:addKeyframe(keyframe)
                    hasFrames = true

                    j = j + 1
                end

                if hasFrames then
                    self.numAnimationParts = self.numAnimationParts + 1
                    self.animationParts[self.numAnimationParts] = part
                end
            end

            i = i + 1
        end

        self.signals = {}
        self.numSignals = 0

        i = 0
        while true do
            local signalKey = string.format("%s.signals.signal(%d)", key, i)
            if not hasXMLProperty(xmlFile, signalKey) then
                break
            end

            local numSignalLights = 1
            local hasAlternatingLights = Utils.getNoNil(getXMLBool(xmlFile, signalKey .. "#alternatingLights"))

            local j = 0
            while true do
                local lightKey = string.format("%s.light(%d)", signalKey, j)
                if not hasXMLProperty(xmlFile, lightKey) then
                    break
                end

                local light = {
                    intensity = 0,
                    offsetValue = 0,
                    isAlternatingLight = hasAlternatingLights
                }

                local shaderNode = I3DUtil.indexToObject(nodeId, getXMLString(xmlFile, lightKey .. "#shaderNode"))

                if shaderNode ~= nil then
                    light.shaderNode = shaderNode
                    light.useLightControl = getHasShaderParameter(light.shaderNode, "lightControl")

                    if light.useLightControl then
                        local _, y, z, w = getShaderParameter(light.shaderNode, "lightControl")

                        light.intensity = math.max(Utils.getNoNil(getXMLFloat(xmlFile, lightKey .. "#intensity"), 15), 0.1)
                        setShaderParameter(light.shaderNode, "lightControl", 0, y, z, w, false)
                    else
                        setVisibility(light.shaderNode, false)
                    end
                end

                local realNode = I3DUtil.indexToObject(nodeId, getXMLString(xmlFile, lightKey .. "#realNode"))

                if realNode ~= nil then
                    light.realNode = realNode

                    if useRealLights then
                        if getHasClassId(light.realNode, ClassIds.LIGHT_SOURCE) then
                            light.colour = {getLightColor(light.realNode)}
                            setVisibility(light.realNode, false)
                        else
                            light.realNode = nil
                        end
                    else
                        setVisibility(light.realNode, false)
                        light.realNode = nil
                    end
                end

                if hasAlternatingLights then
                    if numSignalLights % 2 == 0 then
                        if light.shaderNode ~= nil and getHasShaderParameter(light.shaderNode, "blinkOffset") then
                            setShaderParameter(light.shaderNode, "blinkOffset", 0.5, 0, 0, 0, false)
                        end

                        if light.realNode ~= nil then
                            setLightColor(light.realNode, light.colour[1] * 0.2, light.colour[2] * 0.2, light.colour[3] * 0.2)
                        end

                        light.offsetValue = math.pi
                    end

                    if light.realNode ~= nil then
                        table.insert(self.alternatingLights, light)
                    end
                end

                numSignalLights = numSignalLights + 1

                table.insert(self.lights, light)

                j = j + 1
            end

            i = i + 1
        end

        self.numLights = #self.lights
        self.numAlternatingLights = #self.alternatingLights

        if self.isClient then
            self.samples = {}

            i = 0
            while true do
                local sample = g_soundManager:loadSampleFromXML(xmlFile, key .. ".samples", string.format("sample(%d)", i), self.baseDirectory, nodeId, 0, AudioGroup.ENVIRONMENT, nil, self)
                if sample == nil then
                    break
                end

                table.insert(self.samples, sample)

                i = i + 1
            end
        end

        return true
    else
        g_logManager:xmlWarning(xmlFilename, "Required trigger at '%s.trigger' could not be found!", key)
    end

    return false
end

function RailroadBarrier:delete()
    self.active = false
    self.requiresUpdate = false

    self.animationDirection = 0
    self.numAnimationParts = 0

    if self.triggerId ~= nil then
        removeTrigger(self.triggerId)

        for i = 0, getNumOfChildren(self.triggerId) - 1 do
            removeTrigger(getChildAt(self.triggerId, i))
        end

        self.triggerId = nil
    end

    if self.isClient and self.samples ~= nil then
        g_soundManager:deleteSamples(self.samples)
        self.samples = nil
    end
end

function RailroadBarrier:loadFrameValues(node, xmlFile, key)
    local dx, dy, dz = getTranslation(node)
    local x, y, z = StringUtil.getVectorFromString(getXMLString(xmlFile, key .. "#translation"))

    x = Utils.getNoNil(x, dx)
    y = Utils.getNoNil(y, dy)
    z = Utils.getNoNil(z, dz)

    local drx, dry, drz = getRotation(node)
    local rx, ry, rz = StringUtil.getVectorFromString(getXMLString(xmlFile, key .. "#rotation"))

    rx = Utils.getNoNilRad(rx, drx)
    ry = Utils.getNoNilRad(ry, dry)
    rz = Utils.getNoNilRad(rz, drz)

    local dsx, dsy, dsz = getScale(node)
    local sx, sy, sz = StringUtil.getVectorFromString(getXMLString(xmlFile, key .. "#scale"))

    sx = Utils.getNoNil(sx, dsx)
    sy = Utils.getNoNil(sy, dsy)
    sz = Utils.getNoNil(sz, dsz)

    local isVisible = Utils.getNoNil(getXMLBool(xmlFile, key .. "#visibility"), true)
    local visibility = isVisible and 1 or 0

    return x, y, z, rx, ry, rz, sx, sy, sz, visibility
end

function RailroadBarrier:update(dt)
    if self.requiresUpdate then
        self.requiresUpdate = false

        self.activeTriggerTime = self.activeTriggerTime + dt

        if self.numAnimationParts > 0 and self.animationDirection ~= 0 then
            local newAnimationTime = MathUtil.clamp(self.animationTime + self.animationDirection * dt / self.animationDuration, 0, 1)

            for _, part in pairs(self.animationParts) do
                local v = part.animCurve:get(newAnimationTime)

                setTranslation(part.node, v[1], v[2], v[3])
                setRotation(part.node, v[4], v[5], v[6])
                setScale(part.node, v[7], v[8], v[9])
                setVisibility(part.node, v[10] == 1)
            end

            self.animationTime = newAnimationTime

            if newAnimationTime == 0 or newAnimationTime == 1 then
                self.animationDirection = 0
            end

            self.requiresUpdate = true
        end

        -- Sync real lights with shader 'blinkOffset' using cosinus
        if self.numAlternatingLights > 0 and self.active then
            for _, light in pairs(self.alternatingLights) do
                if light.realNode ~= nil then
                    local alpha = MathUtil.clamp(math.cos(7 * getShaderTimeSec() + light.offsetValue) + 0.2, 0, 1)
                    setLightColor(light.realNode, light.colour[1] * alpha, light.colour[2] * alpha, light.colour[3] * alpha)
                end
            end

            self.requiresUpdate = true
        end

        if self.activeTriggerTime > self.objectCheckTime then
            self.objectCheckTime = self.activeTriggerTime + RailroadBarrier.CHECK_TIME

            for otherId, _ in pairs (self.objectsInTrigger) do
                if not entityExists(otherId) then
                    self.objectsInTrigger[otherId] = nil
                end
            end

            self.numObjectsInTrigger = ListUtil.size(self.objectsInTrigger)
            self:setActive(self.numObjectsInTrigger ~= 0)
        end

        if RailroadBarrier.DEBUG_ACTIVE then
            local secondsString = string.format("%02d", (self.activeTriggerTime / 1000) % 60)

            g_currentMission:addExtraPrintText(string.format("Active trigger time (%s): %02d:%s", self.identifierName, math.floor(self.activeTriggerTime / 60000), secondsString))
            g_currentMission:addExtraPrintText(string.format("Objects in trigger (%s): %d", self.identifierName, self.numObjectsInTrigger))
        end
    end
end

function RailroadBarrier:setActive(active)
    if self.active ~= active then
        self.active = Utils.getNoNil(active, false)

        if self.numLights > 0 then
            for _, light in pairs(self.lights) do
                if light.shaderNode ~= nil then
                    if light.useLightControl then
                        local _, y, z, w = getShaderParameter(light.shaderNode, "lightControl")
                        setShaderParameter(light.shaderNode, "lightControl", self.active and light.intensity or 0, y, z, w, false)
                    else
                        setVisibility(light.shaderNode, self.active)
                    end
                end

                if light.realNode ~= nil then
                    setVisibility(light.realNode, self.active)
                end
            end
        end

        if self.active then
            if self.isClient and self.samples ~= nil then
                if not self.samplesActive then
                    g_soundManager:playSamples(self.samples)
                    self.samplesActive = true
                end
            end

            self.animationDirection = 1
        else
            if self.isClient and self.samples ~= nil then
                if self.samplesActive then
                    g_soundManager:stopSamples(self.samples)
                    self.samplesActive = false
                end
            end

            self.animationDirection = -1
        end
    end

    self.requiresUpdate = true
end

function RailroadBarrier:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
    if (onEnter or onLeave) and otherId ~= 0 then
        if onEnter then
            if self.objectsInTrigger[otherId] == nil then
                self.numObjectsInTrigger = self.numObjectsInTrigger + 1
                self.objectsInTrigger[otherId] = self.numObjectsInTrigger

                if self.numObjectsInTrigger == 1 then
                    self.activeTriggerTime = 0
                    self.objectCheckTime = RailroadBarrier.CHECK_TIME
                end

                self:setActive(true)
            end
        else
            if self.objectsInTrigger[otherId] ~= nil then
                self.numObjectsInTrigger = math.max(self.numObjectsInTrigger - 1, 0)
                self.objectsInTrigger[otherId] = nil

                self:setActive(self.numObjectsInTrigger ~= 0)
            end
        end
    end
end

if g_modManager ~= nil and g_modManager:isModMap(g_currentModName) then
    print("  Register modOnCreate: RailroadBarrier")
    g_onCreateUtil.addOnCreateFunction("RailroadBarrier", RailroadBarrier.onCreate)
end
